/* ppng.c */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <png.h>
#include "print.h"

/* these functions are from print.c */

/* registers printer with printer driver */
extern void registerPrinter(getPrinterInfoProc getPrtInfo,
 initPrintProc initPrt, printBandProc prtBand, endPrintProc endPrt);

/* put the files in the print queue of the currently selected printer */
extern int enqueuePrintFile(char *file1);

/* cleanup of temporary files */
extern void removeFileFreeFilename(char *filename);

static FILE *fp;
static char *fname;
static int printModeType;
static png_structp png_ptr;
static png_infop info_ptr;
static int xpix, ypix;

static int initPrint(int printType, int complexImage, int xpixels,
 int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 png_uint_32 res_x, res_y;

 printModeType = printType;

 fname = tempnam("/tmp","prt");
 if (fname==NULL)
  return(PRT_IO_ERROR);

 if ((fp = fopen(fname, "wb")) == NULL)
  return(PRT_IO_ERROR);

 png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
    (void *) NULL, NULL, NULL);
 if (!png_ptr)
  { fclose(fp); return(PRT_NO_MEMORY); }

 info_ptr = png_create_info_struct(png_ptr);
 if (!info_ptr)
  {
   fclose(fp);
   png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
   return(PRT_NO_MEMORY);
  }

 if (setjmp(png_ptr->jmpbuf))
  {
   fclose(fp);
   png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
   return(PRT_IO_ERROR);
  }

 png_init_io(png_ptr, fp);

 /* need to delay second part of initialisation until we get the colormap */
 xpix = xpixels;
 ypix = ypixels;
 res_x = (png_uint_32) (xdpi/25.4 + 0.5);
 res_y = (png_uint_32) (ydpi/25.4 + 0.5);
 png_set_pHYs(png_ptr, info_ptr, res_x, res_y, PNG_RESOLUTION_METER);
 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
 int lines, char bandType, unsigned char *cmap)
{
 int x, y;

 if (data==NULL)
  return(PRT_INTERNAL_ERROR);
 if (printModeType<bandType)
  return(PRT_WRONG_IMG);

 if (ypos==0)
  {
   int i;
   png_colorp palette;

   switch (printModeType)
    {
     case IMG_BW:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 1, PNG_COLOR_TYPE_GRAY,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      break;
     case IMG_CMAP:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 8, PNG_COLOR_TYPE_PALETTE,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      palette = (png_colorp)png_malloc(png_ptr, 256 * sizeof(png_color));
      for (i = 0;  i < 256;  i++)
       {
	palette[i].red = cmap[i*3];
	palette[i].green = cmap[i*3+1];
	palette[i].blue = cmap[i*3+2];
       }
      png_set_PLTE(png_ptr, info_ptr, palette, 256);
      break;
     case IMG_TRUE_COL:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 8, PNG_COLOR_TYPE_RGB,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      break;
    }
   png_write_info(png_ptr, info_ptr);
   png_set_invert_mono(png_ptr);
  }
 
 switch (printModeType)
  {
   case IMG_BW:
    for (y=0; y<lines; y++)
     {
      png_bytep row_pointers = &data[y*bpl];
      png_write_rows(png_ptr, &row_pointers, 1);
     }
    break;
   case IMG_CMAP:
    if (bandType==IMG_BW)
     {
      unsigned char *dat, cdata[bpl*8];
      png_bytep row_pointers = (png_bytep) &cdata;

      for (y=0; y<lines; y++)
       {
        dat = &data[y*bpl];
        for (x=0; x<bpl*8; x++)
	 cdata[x] = (dat[x>>3]&(128>>(x&7))) ? 255 : 254;
        png_write_rows(png_ptr, &row_pointers, 1);
       }
     }
    else /* (bandType==IMG_CMAP) */
     for (y=0; y<lines; y++)
      {
       png_bytep row_pointers = &data[y*bpl];
       png_write_rows(png_ptr, &row_pointers, 1);
      }
    break;
   case IMG_TRUE_COL:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = &data[y*bpl];
	  memset(cdata, 0, bpl*8*3);
          for (x=0; x<bpl*8; x++)
	   if (!(dat[x>>3]&(128>>(x&7))))
	    { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; }
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = &data[y*bpl];
          for (x=0; x<bpl; x++)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2];}
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
       for (y=0; y<lines; y++)
	{
	 png_bytep row_pointers = &data[y*bpl];
	 png_write_rows(png_ptr, &row_pointers, 1);
	}
       break;
     }
    break;
  } 
 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 png_write_end(png_ptr, info_ptr);

 if (printModeType==IMG_CMAP)
  free(info_ptr->palette);
 png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
 fclose(fp);
 if (printStatus==PRT_OK)
  printStatus = enqueuePrintFile(fname);

 /* just cleanup; called even if printing fails */
 removeFileFreeFilename(fname);
 return(printStatus);
}

/* fill in the preallocated printer modes for Primera */
static void getPrtInfo(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_CMAP;
 printerMod[1].printType = IMG_TRUE_COL;
 printerMod[2].printType = IMG_BW;
 *numModes=3;

 *numDensitX = *numDensitY = 6;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = 20000;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=TRUE;
 *prtName="PNG";
 *profileName=NULL;
}

/* register PNG "printer" */
void registerPNG()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
