/* ptiff.c */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>
#include "print.h"

/* these functions are from print.c */

/* registers printer with printer driver */
extern void registerPrinter(getPrinterInfoProc getPrtInfo,
 initPrintProc initPrt, printBandProc prtBand, endPrintProc endPrt);

/* put the files in the print queue of the currently selected printer */
extern int enqueuePrintFile(char *file1);

/* cleanup of temporary files */
extern void removeFileFreeFilename(char *filename);

static char *fname;
static int printModeType;
static TIFF *out;

static int initPrint(int printType, int complexImage, int xpixels,
	int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 int samplesPP=1, bitsPP=1;
 unsigned short photometric=0;

 printModeType = printType;

 out = NULL;
 switch (printModeType)
  {
   case IMG_BW:
    photometric = PHOTOMETRIC_MINISBLACK;
    samplesPP = 1;
    bitsPP = 1;
    break;
   case IMG_CMAP:
    photometric = PHOTOMETRIC_PALETTE;
    samplesPP = 1;
    bitsPP = 8;
    break;
   case IMG_TRUE_COL:
    photometric = PHOTOMETRIC_RGB;
    samplesPP = 3;
    bitsPP = 8;
    break;
  }

 fname = tempnam("/tmp","prt");
 if (fname==NULL)
  return(PRT_IO_ERROR);

 if ((out = TIFFOpen(fname, "w")) == NULL)
  return(PRT_IO_ERROR);

 TIFFSetField(out, TIFFTAG_IMAGEWIDTH, (unsigned long) xpixels);
 TIFFSetField(out, TIFFTAG_IMAGELENGTH, (unsigned long) ypixels);
 TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
 TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, samplesPP);
 TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, bitsPP);
 TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
 TIFFSetField(out, TIFFTAG_PHOTOMETRIC, photometric);
 TIFFSetField(out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
 TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, 1);
 TIFFSetField(out, TIFFTAG_SOFTWARE, "Arahne");
 TIFFSetField(out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
 TIFFSetField(out, TIFFTAG_XRESOLUTION, xdpi/1000.0);
 TIFFSetField(out, TIFFTAG_YRESOLUTION, ydpi/1000.0);
 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
 int lines, char bandType, unsigned char *cmap)
{
 int x, y;

 if (data==NULL)
  return(PRT_INTERNAL_ERROR);
 if (printModeType<bandType)
  return(PRT_WRONG_IMG);

 if ((ypos==0)&&(printModeType == IMG_CMAP))
  {
   unsigned short *r, *g, *b;
   int i;

   r = (unsigned short *) malloc(sizeof(unsigned short)*3*256);
   g = r + 256;
   b = g + 256;
   for (i = 0; i < 256; i++)
    {
     r[i] = cmap[i*3]*256;
     g[i] = cmap[i*3+1]*256;
     b[i] = cmap[i*3+2]*256;
    }
   TIFFSetField(out, TIFFTAG_COLORMAP, r, g, b);
   free(r);
  }

 if ((bandType==IMG_BW)&&(printModeType==IMG_BW))
  for (y=lines*bpl-1; y>=0; y--)
   data[y]=~data[y];
 
 switch (printModeType)
  {
   case IMG_BW:
    for (y=0; y<lines; y++)
      TIFFWriteScanline(out, &data[y*bpl], y+ypos, 0);
    break;
   case IMG_CMAP:
    if (bandType==IMG_BW)
     {
      unsigned char *dat, cdata[bpl*8];

      for (y=0; y<lines; y++)
       {
        dat = &data[y*bpl];
        for (x=0; x<bpl*8; x++)
	 cdata[x] = (dat[x>>3]&(128>>(x&7))) ? 255 : 254;
        TIFFWriteScanline(out,  cdata, y+ypos, 0);
       }
     }
    else /* (bandType==IMG_CMAP) */
    for (y=0; y<lines; y++)
     TIFFWriteScanline(out,  &data[y*bpl], y+ypos, 0);
    break;
   case IMG_TRUE_COL:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = &data[y*bpl];
	  memset(cdata, 0, bpl*8*3);
          for (x=0; x<bpl*8; x++)
	   if (!(dat[x>>3]&(128>>(x&7))))
	    { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; }
          TIFFWriteScanline(out,  cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = &data[y*bpl];
          for (x=0; x<bpl; x++)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2];}
          TIFFWriteScanline(out,  cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
       for (y=0; y<lines; y++)
	TIFFWriteScanline(out, &data[y*bpl], y+ypos, 0);
       break;
     }
    break;
  } 
 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 (void) TIFFClose(out);
 if (printStatus==PRT_OK)
  printStatus = enqueuePrintFile(fname);

 /* just cleanup; called even if printing fails */
 removeFileFreeFilename(fname);
 return(printStatus);
}

/* fill in the preallocated printer modes for Primera */
static void getPrtInfo(printerMode *printerMod, int *numModes,
	printerDensity *prtDensityX, int *numDensitX,
	printerDensity *prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_CMAP;
 printerMod[1].printType = IMG_TRUE_COL;
 printerMod[2].printType = IMG_BW;

 *numModes=3;

 *numDensitX = *numDensitY = 6;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = 20000;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=TRUE;
 *prtName="TIFF";
 *profileName=NULL;
}

/* register TIFF "printer" */
void registerTIFF()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
